"""
``python_tsp.exact``
====================

Module with exact solvers, which have a guarantee of returning the optimal
solution of a TSP problem. They may be good options for small scale problems
that need to be solved many times in a row, even more if when calling other
solvers can add too much of an overhead in implementing a TSP model, calling a
wrapper from another language etc.

It is up to the user to decide how small enough is a "small scale problem". We
suggest to test the solvers with toy problems with typical number of nodes
required and check if the processing time is acceptable.

Functions present in this module are listed below.

    solve_tsp_brute_force
    solve_tsp_dynamic_programming


Tips on deciding the solver
---------------------------

In general, ``solve_tsp_dynamic_programming`` is faster, specially with an
appropriate `maxsize` input (the default is fine). However, because of its
recursion, it may take more memory, particularly if the number of nodes grows
large. If that becomes an issue and you still need a provably optimal solution,
use the ``solve_tsp_brute_force``.
"""

from .brute_force import solve_tsp_brute_force  # noqa
from .dynamic_programming import solve_tsp_dynamic_programming  # noqa



import logging
from timeit import default_timer
from typing import List, Optional, Tuple

import numpy as np

from python_tsp.heuristics.perturbation_schemes import neighborhood_gen
from python_tsp.heuristics.local_search import setup
from python_tsp.utils import compute_permutation_distance


logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(level=logging.WARNING)
logger.addHandler(ch)


def solve_tsp_simulated_annealing(
    distance_matrix: np.ndarray,
    x0: Optional[List[int]] = None,
    perturbation_scheme: str = "two_opt",
    alpha: float = 0.9,
    max_processing_time: float = None,
    log_file: Optional[str] = None,
) -> Tuple[List, float]:
    """Solve a TSP problem using a Simulated Annealing
    The approach used here is the one proposed in [1].

    Parameters
    ----------
    distance_matrix
        Distance matrix of shape (n x n) with the (i, j) entry indicating the
        distance from node i to j

    x0
        Initial permutation. If not provided, it starts with a random path

    perturbation_scheme {"ps1", "ps2", "ps3", "ps4", "ps5", "ps6", ["two_opt"]}
        Mechanism used to generate new solutions. Defaults to "two_opt"

    alpha
        Reduction factor (``alpha`` < 1) used to reduce the temperature. As a
        rule of thumb, 0.99 takes longer but may return better solutions, whike
        0.9 is faster but may not be as good. A good approach is to use 0.9
        (as default) and if required run the returned solution with a local
        search.

    max_processing_time {None}
        Maximum processing time in seconds. If not provided, the method stops
        only when there were 3 temperature cycles with no improvement.

    log_file
        If not `None`, creates a log file with details about the whole
        execution

    Returns
    -------
    A permutation of nodes from 0 to n - 1 that produces the least total
    distance obtained (not necessarily optimal).

    The total distance the returned permutation produces.

    References
    ----------
    [1] Dréo, Johann, et al. Metaheuristics for hard optimization: methods and
    case studies. Springer Science & Business Media, 2006.
    """

    x, fx = setup(distance_matrix, x0)
    temp = _initial_temperature(distance_matrix, x, fx, perturbation_scheme)
    max_processing_time = max_processing_time or np.inf
    if log_file:
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.INFO)
        logger.addHandler(fh)
        logger.setLevel(logging.INFO)

    n = len(x)
    k_inner_min = n  # min inner iterations
    k_inner_max = 10 * n  # max inner iterations
    k_noimprovements = 0  # number of inner loops without improvement

    tic = default_timer()
    stop_early = False
    while (k_noimprovements < 3) and (not stop_early):
        k_accepted = 0  # number of accepted perturbations
        for k in range(k_inner_max):
            if default_timer() - tic > max_processing_time:
                logger.warning("Stopping early due to time constraints")
                stop_early = True
                break

            xn = _perturbation(x, perturbation_scheme)
            fn = compute_permutation_distance(distance_matrix, xn)

            if _acceptance_rule(fx, fn, temp):
                x, fx = xn, fn
                k_accepted += 1
                k_noimprovements = 0

            logger.info(
                f"Temperature {temp}. Current value: {fx} "
                f"k: {k + 1}/{k_inner_max} "
                f"k_accepted: {k_accepted}/{k_inner_min} "
                f"k_noimprovements: {k_noimprovements}"
            )

            if k_accepted >= k_inner_min:
                break

        temp *= alpha  # temperature update
        k_noimprovements += k_accepted == 0

    return x, fx


def _initial_temperature(
    distance_matrix: np.ndarray,
    x: List[int],
    fx: float,
    perturbation_scheme: str,
) -> float:
    """Compute initial temperature
    Instead of relying on problem-dependent parameters, this function estimates
    the temperature using the suggestion in [1].

    Notes
    -----
    Here are the steps followed:
        1. Generate 100 disturbances at random from T0, and evaluate the mean
        objective value differences dfx_mean = mean(fn - fx);
        2. Choose tau0 = 0.5 as assumed quality of initial solution (assuming
        a bad one), and deduce T0 from exp(-fx_mean/T0) = tau0, that is,
        T0 = -fx_mean/ln(tau0)

    References
    ----------
    [1] Dréo, Johann, et al. Metaheuristics for hard optimization: methods and
    case studies. Springer Science & Business Media, 2006.
    """

    # Step 1
    dfx_list = []
    for _ in range(100):
        xn = _perturbation(x, perturbation_scheme)
        fn = compute_permutation_distance(distance_matrix, xn)
        dfx_list.append(fn - fx)

    dfx_mean = np.abs(np.mean(dfx_list))

    # Step 2
    tau0 = 0.5
    return -dfx_mean / np.log(tau0)


def _perturbation(x: List[int], perturbation_scheme: str):
    """Generate a random neighbor of a current solution ``x``
    In this case, we can use the generators created in the `local_search`
    module, and pick the first solution. Since the neighborhood is randomized,
    it is the same as creating a random perturbation.
    """
    return next(neighborhood_gen[perturbation_scheme](x))


def _acceptance_rule(fx: float, fn: float, temp: float) -> bool:
    """Metropolis acceptance rule"""

    dfx = fn - fx
    return (dfx < 0) or (
        (dfx > 0) and (np.random.rand() <= np.exp(-(fn - fx) / temp))
    )



from functools import lru_cache
from typing import Dict, List, Optional, Tuple

import numpy as np


def solve_tsp_dynamic_programming(
    distance_matrix: np.ndarray,
    maxsize: Optional[int] = None
) -> Tuple[List, float]:
    """
    Solve TSP to optimality with dynamic programming.

    Parameters
    ----------
    distance_matrix
        Distance matrix of shape (n x n) with the (i, j) entry indicating the
        distance from node i to j. It does not need to be symmetric

    maxsize
        Parameter passed to ``lru_cache`` decorator. Used to define the maximum
        size for the recursion tree. Defaults to `None`, which essentially
        means "take as much space as needed".

    Returns
    -------
    permutation
        A permutation of nodes from 0 to n that produces the least total
        distance

    distance
        The total distance the optimal permutation produces

    Notes
    -----
    Algorithm: cost of the optimal path
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Consider a TSP instance with 3 nodes: {0, 1, 2}. Let dist(0, {1, 2}) be the
    distance from 0, visiting all nodes in {1, 2} and going back to 0. This can
    be computed recursively as:

        dist(0, {1, 2}) = min(
            c_{0, 1} + dist(1, {2}),
            c_{0, 2} + dist(2, {1}),
        )

    wherein c_{0, 1} is the cost from going from 0 to 1 in the distance matrix.
    The inner dist(1, {2}) is computed as:

        dist(1, {2}) = min(
            c_{1, 2} + dist(2, {}),
        )

    and similarly for dist(2, {1}). The stopping point in the recursion is:

        dist(2, {}) = c_{2, 0}.

    This process can be generalized as:

        dist(ni, N) =   min   ( c_{ni, nj} + dist(nj, N - {nj}) )
                      nj in N

    and

        dist(ni, {}) = c_{ni, 0}

    With starting point as dist(0, {1, 2, ..., tsp_size}). The notation
    N - {nj} is the difference operator, meaning set N without node nj.


    Algorithm: compute the optimal path
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The previous process returns the distance of the optimal path. To find the
    actual path, we need to store in a memory the following key/values:

        memo[(ni, N)] = nj_min

    with nj_min the node in N that provided the smallest value of dist(ni, N).
    Then, the process goes backwards starting from
    memo[(0, {1, 2, ..., tsp_size})].

    In the previous example, suppose memo[(0, {1, 2})] = 1.
    Then, look for memo[(1, {2})] = 2.
    Then, since the next step would be memo[2, {}], stop there. The optimal
    path would be 0 -> 1 -> 2 -> 0.


    Reference
    ---------
    https://en.wikipedia.org/wiki/Held%E2%80%93Karp_algorithm#cite_note-5
    """
    # Get initial set {1, 2, ..., tsp_size} as a frozenset because since
    # @lru_cache requires a hashable type
    N = frozenset(range(1, distance_matrix.shape[0]))
    memo: Dict[Tuple, int] = {}

    # Step 1: get minimum distance
    @lru_cache(maxsize=maxsize)
    def dist(ni: int, N: frozenset) -> float:
        if not N:
            return distance_matrix[ni, 0]

        # Store the costs in the form (nj, dist(nj, N))
        costs = [
            (nj, distance_matrix[ni, nj] + dist(nj, N.difference({nj})))
            for nj in N
        ]
        nmin, min_cost = min(costs, key=lambda x: x[1])
        memo[(ni, N)] = nmin
        return min_cost

    best_distance = dist(0, N)

    # Step 2: get path with the minimum distance
    ni = 0  # start at the origin
    solution = [0]
    while N:
        ni = memo[(ni, N)]
        solution.append(ni)
        N = N.difference({ni})

    return solution, best_distance



"""Contains typical distance matrices"""
from typing import List, Optional, Tuple

import numpy as np

EARTH_RADIUS_METERS = 6371000


def euclidean_distance_matrix(
    sources: np.ndarray, destinations: Optional[np.ndarray] = None
) -> np.ndarray:
    """Distance matrix using the Euclidean distance

    Parameters
    ----------
    sources, destinations
        Arrays with each row containing the coordinates of a point. If
        ``destinations`` is None, compute the distance between each source in
        ``sources``.

    Returns
    -------
    distance_matrix
        Array with the (i. j) entry indicating the Euclidean distance between
        the i-th row in `sources` and the j-th row in `destinations`.

    Notes
    -----
    The Euclidean distance between points x = (x1, x2, ..., xn) and
    y = (y1, y2, ..., yn), with n coordinates each, is given by:

        sqrt((y1 - x1)**2 + (y2 - x2)**2 + ... + (yn - xn)**2)

    If the user requires the distance between each point in a single array,
    call this this function with `sources` = `destinations`.
    """
    sources, destinations = _process_input(sources, destinations)
    return np.sqrt(
        ((sources[:, :, None] - destinations[:, :, None].T) ** 2).sum(1)
    )


def great_circle_distance_matrix(
    sources: np.ndarray, destinations: Optional[np.ndarray] = None
) -> np.ndarray:
    """Distance matrix using the Great Circle distance
    This is an Euclidean-like distance but on spheres [1]. In this case it is
    used to estimate the distance in meters between locations in the Earth.

    Parameters
    ----------
    sources, destinations
        Arrays with each row containing the coordinates of a point in the form
        [lat, lng]. Notice it only considers the first two columns.
        Also, if ``destinations`` is `None`, compute the distance between each
        source in ``sources``.

    Returns
    -------
    distance_matrix
        Array with the (i. j) entry indicating the Great Circle distance (in
        meters) between the i-th row in `sources` and the j-th row in
        `destinations`.

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Great-circle_distance
    Using the third computational formula
    """

    sources, destinations = _process_input(sources, destinations)
    sources_rad = np.radians(sources)
    dests_rad = np.radians(destinations)

    delta_lambda = sources_rad[:, [1]] - dests_rad[:, 1]  # (N x M) lng
    phi1 = sources_rad[:, [0]]  # (N x 1) array of source latitudes
    phi2 = dests_rad[:, 0]  # (1 x M) array of destination latitudes

    delta_sigma = np.arctan2(
        np.sqrt(
            (np.cos(phi2) * np.sin(delta_lambda))**2 +
            (np.cos(phi1) * np.sin(phi2) -
             np.sin(phi1) * np.cos(phi2) * np.cos(delta_lambda))**2
        ),
        (np.sin(phi1) * np.sin(phi2) +
         np.cos(phi1) * np.cos(phi2) * np.cos(delta_lambda))
    )

    return EARTH_RADIUS_METERS * delta_sigma


def _process_input(
    sources: np.ndarray, destinations: Optional[np.ndarray] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Pre-process input
    This function ensures ``sources`` and ``destinations`` have at least two
    dimensions, and if ``destinations`` is `None`, set it equal to ``sources``.
    """
    if destinations is None:
        destinations = sources

    sources = np.atleast_2d(sources)
    destinations = np.atleast_2d(destinations)

    return sources, destinations


def tsplib_distance_matrix(tsplib_file: str) -> np.ndarray:
    """Distance matrix from a TSPLIB file

    Parameters
    ----------
    tsplib_file
        A string with the complete path of the TSPLIB file (or just its name if
        it is the in current path)

    Returns
    -------
    distance_matrix
        A ND-array with the equivalent distance matrix of the input file

    Raises
    ------
    NotImplementedError
        In case a not supported file is provided.

    Notes
    -----
    This function is able to handle the following types of TSPLIB files:
        - EDGE_WEIGHT_TYPE: "EUC_2D", "CEIL_", "GEO" and "EXPLICIT"
        - EDGE_WEIGHT_FORMAT: "FULL_MATRIX", "LOWER_ROW", "LOWER_DIAG_ROW",
        "UPPER_ROW" and "UPPER_DIAG_ROW".
    """
    with open(tsplib_file, "r") as f:
        lines = f.readlines()

    # Get file relevant attributes
    edge_weigth_type = _find_file_attribute(lines, "EDGE_WEIGHT_TYPE")
    edge_weigth_format = _find_file_attribute(lines, "EDGE_WEIGHT_FORMAT")

    if edge_weigth_type in ("EUC_2D", "CEIL_2D"):
        return _euc_2d_tsplib_distance_matrix(lines, edge_weigth_type)
    elif edge_weigth_type == "GEO":
        return _geo_tsplib_distance_matrix(lines)
    elif (
        edge_weigth_type == "EXPLICIT" and edge_weigth_format == "FULL_MATRIX"
    ):
        return _explicit_full_matrix_tsplib_distance_matrix(lines)
    elif (
        edge_weigth_type == "EXPLICIT"
        and edge_weigth_format in ("LOWER_ROW", "LOWER_DIAG_ROW")
    ):
        offset = 0 if edge_weigth_format == "LOWER_DIAG_ROW" else 1
        return _explicit_lower_row_tsplib_distance_matrix(lines, offset=offset)
    elif (
        edge_weigth_type == "EXPLICIT"
        and edge_weigth_format in ("UPPER_ROW", "UPPER_DIAG_ROW")
    ):
        offset = 0 if edge_weigth_format == "UPPER_DIAG_ROW" else 1
        return _explicit_upper_row_tsplib_distance_matrix(lines, offset=offset)
    else:
        raise NotImplementedError("tsplib file not supported")


def _find_file_attribute(lines: List[str], attribute: str) -> str:
    line = next((line for line in lines if line.startswith(attribute)), None)
    if line:
        return line.split(":")[1].strip()
    return ""


def _euc_2d_tsplib_distance_matrix(
    lines: List[str], edge_weigth_type: str
) -> np.ndarray:

    coordinates = _get_coordinates(lines)

    distance_matrix = euclidean_distance_matrix(coordinates)
    if edge_weigth_type == "CEIL_2D":
        distance_matrix = np.ceil(distance_matrix)

    return distance_matrix.astype(int)


def _geo_tsplib_distance_matrix(lines: List[str]) -> np.ndarray:

    coordinates = _get_coordinates(lines)
    distance_matrix = great_circle_distance_matrix(coordinates)

    return distance_matrix.astype(int)


def _get_coordinates(lines: List[str]) -> np.ndarray:
    """Get the coordinates in case of files without an explicit matrix"""
    dimension = int(_find_file_attribute(lines, "DIMENSION"))
    # Get the index of the line starting with the nodes information
    node_section_index = next(
        (
            i for i, line in enumerate(lines)
            if line.startswith("NODE_COORD_SECTION")
        )
    ) + 1

    def read_node_coordinates(line):
        return [float(coordinate) for coordinate in line.split()[1:]]

    return np.array([
        read_node_coordinates(line)
        for line in lines[node_section_index:node_section_index + dimension]
    ])


def _explicit_full_matrix_tsplib_distance_matrix(
    lines: List[str]
) -> np.ndarray:
    dimension = int(_find_file_attribute(lines, "DIMENSION"))
    distance_matrix_flattened = _get_distance_matrix_flattened(lines)

    distance_matrix = np.reshape(
        distance_matrix_flattened, (dimension, dimension)
    )
    np.fill_diagonal(distance_matrix, 0)
    return distance_matrix


def _explicit_lower_row_tsplib_distance_matrix(
    lines: List[str], offset: int = 0
) -> np.ndarray:
    dimension = int(_find_file_attribute(lines, "DIMENSION"))
    distance_matrix_flattened = _get_distance_matrix_flattened(lines)

    # Construct the lower part of the distance matrix
    distance_matrix = np.zeros((dimension, dimension))
    tril_indices = np.tril_indices(dimension, k=offset, m=dimension)
    distance_matrix[tril_indices] = distance_matrix_flattened
    distance_matrix += distance_matrix.transpose()
    return distance_matrix


def _explicit_upper_row_tsplib_distance_matrix(
    lines: List[str], offset: int = 0
) -> np.ndarray:
    dimension = int(_find_file_attribute(lines, "DIMENSION"))
    distance_matrix_flattened = _get_distance_matrix_flattened(lines)

    # Construct the upper part of the distance matrix
    distance_matrix = np.zeros((dimension, dimension))
    triu_indices = np.triu_indices(dimension, k=offset, m=dimension)
    distance_matrix[triu_indices] = distance_matrix_flattened
    distance_matrix += distance_matrix.transpose()
    return distance_matrix


def _get_distance_matrix_flattened(lines: List[str]) -> List[int]:
    """Return all cells of the distance matrix in the TSPLIB file"""

    # Get the index of the line starting with the edges information
    edge_section_index = next(
        (
            i for i, line in enumerate(lines)
            if line.startswith("EDGE_WEIGHT_SECTION")
        )
    ) + 1

    # Each file line does not necessarily contains one matrix row. Thus, read
    # all matrix data at once first and then reshape the array later
    def read_line_data(line):
        return [int(cell) for cell in line.split()]

    distance_matrix_flattened = []
    for line in lines[edge_section_index:]:
        # Try reading the current line and stop if a non-integer pops up (such
        # as "EOF" or another section)
        try:
            distance_matrix_flattened.extend(read_line_data(line))
        except ValueError:
            break

    return distance_matrix_flattened



"""Module with a brute force TSP solver"""
from itertools import permutations
from typing import Any, List, Optional, Tuple

import numpy as np

from python_tsp.utils import compute_permutation_distance


def solve_tsp_brute_force(
    distance_matrix: np.ndarray
) -> Tuple[Optional[List], Any]:
    """Solve TSP to optimality with a brute force approach

    Parameters
    ----------
    distance_matrix
        Distance matrix of shape (n x n) with the (i, j) entry indicating the
        distance from node i to j. It does not need to be symmetric

    Returns
    -------
    permutation
        A permutation of nodes from 0 to n that produces the least total
        distance

    distance
        The total distance the optimal permutation produces

    Notes
    ----
    The algorithm checks all permutations and returns the one with smallest
    distance. In principle, the total number of possibilities would be n! for
    n nodes. However, we can fix node 0 and permutate only the remaining,
    reducing the possibilities to (n - 1)!.
    """

    # Exclude 0 from the range since it is fixed as starting point
    points = range(1, distance_matrix.shape[0])
    best_distance = np.inf
    best_solution = None
    for partial_permutation in permutations(points):
        # Remember to add the starting node before evaluating it
        permutation = [0] + list(partial_permutation)
        distance = compute_permutation_distance(distance_matrix, permutation)
        if distance < best_distance:
            best_distance = distance
            best_solution = permutation

    return best_solution, best_distance



"""Perturbation schemes used in local search-based algorithms.
The functions here should receive a permutation list with numbers from 0 to `n`
and return a generator with all neighbors of this permutation.
The neighbors should preferably be randomized to be used in Simulated
Annealing, which samples a single neighbor at a time.

References
----------
.. [1] Goulart, Fillipe, et al. "Permutation-based optimization for the load
    restoration problem with improved time estimation of maneuvers."
    International Journal of Electrical Power & Energy Systems 101 (2018):
    339-355.
.. [2] 2-opt: https://en.wikipedia.org/wiki/2-opt
"""
from random import sample
from typing import Callable, Dict, Generator, List


def ps1_gen(x: List[int]) -> Generator[List[int], List[int], None]:
    """PS1 perturbation scheme: Swap two adjacent terms [1]
    This scheme has at most n - 1 swaps.
    """

    n = len(x)
    i_range = range(1, n - 1)
    for i in sample(i_range, len(i_range)):
        xn = x.copy()
        xn[i], xn[i + 1] = xn[i + 1], xn[i]
        yield xn


def ps2_gen(x: List[int]) -> Generator[List[int], List[int], None]:
    """PS2 perturbation scheme: Swap any two elements [1]
    This scheme has n * (n - 1) / 2 swaps.
    """

    n = len(x)
    i_range = range(1, n - 1)
    for i in sample(i_range, len(i_range)):
        j_range = range(i + 1, n)
        for j in sample(j_range, len(j_range)):
            xn = x.copy()
            xn[i], xn[j] = xn[j], xn[i]
            yield xn


def ps3_gen(x: List[int]) -> Generator[List[int], List[int], None]:
    """PS3 perturbation scheme: A single term is moved [1]
    This scheme has n * (n - 1) swaps.
    """

    n = len(x)
    i_range = range(1, n)
    for i in sample(i_range, len(i_range)):
        j_range = [j for j in range(1, n) if j != i]
        for j in sample(j_range, len(j_range)):
            xn = x.copy()
            node = xn.pop(i)
            xn.insert(j, node)
            yield xn


def ps4_gen(x: List[int]) -> Generator[List[int], List[int], None]:
    """PS4 perturbation scheme: A subsequence is moved [1]"""

    n = len(x)
    i_range = range(1, n)
    for i in sample(i_range, len(i_range)):
        j_range = range(i + 1, n + 1)
        for j in sample(j_range, len(j_range)):
            k_range = [k for k in range(1, n + 1) if k not in range(i, j + 1)]
            for k in sample(k_range, len(k_range)):
                xn = x.copy()
                if k < i:
                    xn = xn[:k] + xn[i:j] + xn[k:i] + xn[j:]
                else:
                    xn = xn[:i] + xn[j:k] + xn[i:j] + xn[k:]
                yield xn


def ps5_gen(x: List[int]) -> Generator[List[int], List[int], None]:
    """PS5 perturbation scheme: A subsequence is reversed [1]"""

    n = len(x)
    i_range = range(1, n)
    for i in sample(i_range, len(i_range)):
        j_range = range(i + 2, n + 1)
        for j in sample(j_range, len(j_range)):
            xn = x.copy()
            xn = xn[:i] + list(reversed(xn[i:j])) + xn[j:]
            yield xn


def ps6_gen(x: List[int]) -> Generator[List[int], List[int], None]:
    """PS6 perturbation scheme: A subsequence is reversed and moved [1]"""

    n = len(x)
    i_range = range(1, n)
    for i in sample(i_range, len(i_range)):
        j_range = range(i + 1, n + 1)
        for j in sample(j_range, len(j_range)):
            k_range = [k for k in range(1, n + 1) if k not in range(i, j + 1)]
            for k in sample(k_range, len(k_range)):
                xn = x.copy()
                if k < i:
                    xn = xn[:k] + list(reversed(xn[i:j])) + xn[k:i] + xn[j:]
                else:
                    xn = xn[:i] + xn[j:k] + list(reversed(xn[i:j])) + xn[k:]
                yield xn


def two_opt_gen(x: List[int]) -> Generator[List[int], List[int], None]:
    """2-opt perturbation scheme [2]"""
    n = len(x)
    i_range = range(2, n)
    for i in sample(i_range, len(i_range)):
        j_range = range(i + 1, n + 1)
        for j in sample(j_range, len(j_range)):
            xn = x.copy()
            xn = xn[:i - 1] + list(reversed(xn[i - 1:j])) + xn[j:]
            yield xn


# Mapping with all possible neighborhood generators in this module
neighborhood_gen: Dict[
    str, Callable[[List[int]], Generator[List[int], List[int], None]]
] = {
    "ps1": ps1_gen,
    "ps2": ps2_gen,
    "ps3": ps3_gen,
    "ps4": ps4_gen,
    "ps5": ps5_gen,
    "ps6": ps6_gen,
    "two_opt": two_opt_gen,
}



"""Auxiliar functions that may be used in most modules"""
from typing import List

import numpy as np


def compute_permutation_distance(
    distance_matrix: np.ndarray, permutation: List[int]
) -> float:
    """Compute distance of a given permutation

    Parameters
    ----------
    distance_matrix
        Distance matrix of shape (n x n) with the (i, j) entry indicating the
        distance from node i to j. It does not need to be symmetric

    permutation
        A list with nodes from 0 to n-1 in any order

    Returns
    -------
    Total distance of the path given in `permutation` for the provided
    `distance_matrix`

    Notes
    -----
    Suppose the permutation [0, 1, 2, 3], with four nodes. The total distance
    of this path will be from 0 to 1, 1 to 2, 2 to 3, and 3 back to 0. This
    can be fetched from a distance matrix using:

        distance_matrix[ind1, ind2], where
        ind1 = [0, 1, 2, 3] # the FROM nodes
        ind2 = [1, 2, 3, 0] # the TO nodes

    This can easily be generalized to any permutation by using ind1 as the
    given permutation, and moving the first node to the end to generate ind2.
    """
    ind1 = permutation
    ind2 = permutation[1:] + permutation[:1]
    return distance_matrix[ind1, ind2].sum()



"""Simple local search solver"""
import logging
from random import sample
from timeit import default_timer
from typing import List, Optional, Tuple

import numpy as np

from python_tsp.utils import compute_permutation_distance
from python_tsp.heuristics.perturbation_schemes import neighborhood_gen


logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(level=logging.WARNING)
logger.addHandler(ch)


def solve_tsp_local_search(
    distance_matrix: np.ndarray,
    x0: Optional[List[int]] = None,
    perturbation_scheme: str = "two_opt",
    max_processing_time: Optional[float] = None,
    log_file: Optional[str] = None,
) -> Tuple[List, float]:
    """Solve a TSP problem with a local search heuristic

    Parameters
    ----------
    distance_matrix
        Distance matrix of shape (n x n) with the (i, j) entry indicating the
        distance from node i to j

    x0
        Initial permutation. If not provided, it starts with a random path

    perturbation_scheme {"ps1", "ps2", "ps3", "ps4", "ps5", "ps6", ["two_opt"]}
        Mechanism used to generate new solutions. Defaults to "two_opt"

    max_processing_time {None}
        Maximum processing time in seconds. If not provided, the method stops
        only when a local minimum is obtained

    log_file
        If not `None`, creates a log file with details about the whole
        execution

    Returns
    -------
    A permutation of nodes from 0 to n - 1 that produces the least total
    distance obtained (not necessarily optimal).

    The total distance the returned permutation produces.

    Notes
    -----
    Here are the steps of the algorithm:
        1. Let `x`, `fx` be a initial solution permutation and its objective
        value;
        2. Perform a neighborhood search in `x`:
            2.1 For each `x'` neighbor of `x`, if `fx'` < `fx`, set `x` <- `x'`
            and stop;
        3. Repeat step 2 until all neighbors of `x` are tried and there is no
        improvement. Return `x`, `fx` as solution.
    """
    x, fx = setup(distance_matrix, x0)
    max_processing_time = max_processing_time or np.inf
    if log_file:
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.INFO)
        logger.addHandler(fh)
        logger.setLevel(logging.INFO)

    tic = default_timer()
    stop_early = False
    improvement = True
    while improvement and (not stop_early):
        improvement = False
        for n_index, xn in enumerate(neighborhood_gen[perturbation_scheme](x)):
            if default_timer() - tic > max_processing_time:
                logger.warning("Stopping early due to time constraints")
                stop_early = True
                break

            fn = compute_permutation_distance(distance_matrix, xn)
            logger.info(f"Current value: {fx}; Neighbor: {n_index}")

            if fn < fx:
                improvement = True
                x, fx = xn, fn
                break  # early stop due to first improvement local search

    return x, fx


def setup(
    distance_matrix: np.ndarray, x0: Optional[List] = None
) -> Tuple[List[int], float]:
    """Return initial solution and its objective value

    Parameters
    ----------
    distance_matrix
        Distance matrix of shape (n x n) with the (i, j) entry indicating the
        distance from node i to j

    x0
        Permutation of nodes from 0 to n - 1 indicating the starting solution.
        If not provided, a random list is created.

    Returns
    -------
    x0
        Permutation with initial solution. If ``x0`` was provided, it is the
        same list

    fx0
        Objective value of x0
    """

    if not x0:
        n = distance_matrix.shape[0]  # number of nodes
        x0 = [0] + sample(range(1, n), n - 1)  # ensure 0 is the first node

    fx0 = compute_permutation_distance(distance_matrix, x0)
    return x0, fx0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['python_tsp', 'python_tsp.exact', 'python_tsp.heuristics']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18.5,<2.0.0']

setup_kwargs = {
    'name': 'python-tsp',
    'version': '0.2.0',
    'description': 'Simple library to solve the Traveling Salesperson Problem in pure Python.',
    'long_description': "=================\nPython TSP Solver\n=================\n\n``python-tsp`` is a library written in pure Python for solving typical Traveling\nSalesperson Problems (TSP). It can work with symmetric and asymmetric versions.\n\n\nInstallation\n============\n.. code:: bash\n\n  pip install python-tsp\n\n\nExamples\n========\n\nGiven a distance matrix as a numpy array, it is easy to compute a Hamiltonian\npath with least cost. For instance, to use a Dynamic Programming method:\n\n.. code:: python\n\n   import numpy as np\n   from python_tsp.exact import solve_tsp_dynamic_programming\n\n   distance_matrix = np.array([\n       [0,  5, 4, 10],\n       [5,  0, 8,  5],\n       [4,  8, 0,  3],\n       [10, 5, 3,  0]\n   ])\n   permutation, distance = solve_tsp_dynamic_programming(distance_matrix)\n\nThe solution will be ``[0, 1, 3, 2]``, with total distance 17. Notice it is\nalways a closed path, so after node 2 we go back to 0.\n\nTo solve the same problem with a metaheuristic method:\n\n.. code:: python\n\n   from python_tsp.heuristics import solve_tsp_simulated_annealing\n\n   permutation, distance = solve_tsp_simulated_annealing(distance_matrix) \n\nKeep in mind that, being a metaheuristic, the solution may vary from execution\nto execution, and there is no guarantee of optimality. However, it may be a\nway faster alternative in larger instances.\n\nIf you with for an open TSP version (it is not required to go back to the\norigin), just set all elements of the first column of the distance matrix to\nzero:\n\n.. code:: python\n\n   distance_matrix[:, 0] = 0\n   permutation, distance = solve_tsp_dynamic_programming(distance_matrix)\n\nand in this case we obtain ``[0, 2, 3, 1]``, with distance 12. Notice that in\nthis case the distance matrix is actually asymmetric, and the methods here are\napplicable as well.\n\nThe previous examples assumed you already had a distance matrix. If that is not\nthe case, the ``distances`` module has prepared some functions to compute an \nEuclidean distance matrix or a\n`Great Circle Distance <https://en.wikipedia.org/wiki/Great-circle_distance>`_.\n\nFor example, if you have an array where each row has the latitude and longitude\nof a point,\n\n.. code:: python\n\n   import numpy as np\n   from python_tsp.distances import great_circle_distance_matrix\n\n   sources = np.array([\n       [ 40.73024833, -73.79440675],\n       [ 41.47362495, -73.92783272],\n       [ 41.26591   , -73.21026228],\n       [ 41.3249908 , -73.507788  ]\n   ])\n   distance_matrix = great_circle_distance_matrix(sources)\n\nSee the `project's repository <https://github.com/fillipe-gsm/python-tsp>`_ \nfor more examples and a list of available methods.\n",
    'author': 'Fillipe Goulart',
    'author_email': 'fillipe.gsm@tutanota.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fillipe-gsm/python-tsp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
